__author__ = 'squirrel'

from math import *

def normalize(v, tolerance=0.00001):
    mag2 = sum(n * n for n in v)
    if abs(mag2 - 1.0) > tolerance:
        mag = sqrt(mag2)
        v = tuple(n / mag for n in v)
    return v

class Quaternion:

    @classmethod
    def from_axisangle(cls, v, theta):
        q = Quaternion(1, v[0], v[1], v[2])
        q.theta = theta
        return q


    @property
    def vector(self):
        return normalize(
            [self.x, self.y, self.z]
        )

    @vector.setter
    def vector(self, vector):
        vector = normalize(vector)
        t = acos(self.w)
        self.x = vector[0]*sin(t)
        self.y = vector[1]*sin(t)
        self.z = vector[2]*sin(t)

    @property
    def theta(self):
        return acos(self.w)*2.0

    @theta.setter
    def theta(self, theta):
        theta /= 2
        v = self.vector
        self.w = cos(theta)
        # note the propery setter above, this is just to re-calculate our vector things :)
        self.vector = v

    def __init__(self, w, x, y, z):

        self.w = w
        self.x = x
        self.y = y
        self.z = z

    def multiply(self, q2):
        w1, x1, y1, z1 = self.w, self.x, self.y, self.z
        w2, x2, y2, z2 = q2.w, q2.x, q2.y, q2.z

        return Quaternion(
            w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2,
            x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2,
            y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2,
            z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
        )

    def conjugate(self):
        return Quaternion(
            w=self.w,
            x=-self.x,
            y=-self.y,
            z=-self.z
        )

    def v_multiply(self, v):
        mag = sqrt(sum(n * n for n in v))

        v = self.multiply(
                    Quaternion(w=0, x=v[0], y=v[1], z=v[2])
                ).multiply(
                    self.conjugate()
                ).vector

        return [
            v[0]*mag,
            v[1]*mag,
            v[2]*mag
        ]

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "Quaternion: v=[{},{},{}], theta={}".format(self.x, self.y, self.z, self.theta)